;****************************************************************************
;
;	DTMF Dekoder mit AT Tiny 2313 CPU
;	Es werden die DTMF frequenzen mit hilfe des Goertzel Algorithmus detektiert.
;	Da die 2313 CPU keine eingebauten ADC besitzt wird hier mit hilfe des Analog
;	komparators und eines an den PORTD angeschlossenen DA-Wandlers ein nach 
;	dem Verfahren der succzessiven Approximation arbeitender AD-Wandler emuliert
;
;	Die Aufloesung dieses ADC betraegt 6 Bit (ca. 36 dB Dynamik Umfang). 
;	Die Abtastrate wird mit hilfe von Timer 0 auf 8.012 KHz eingestellt.
;	Das so digitalierte Signalgemisch wird dann mit hilfe des Goertzel Algorithmus gefiltert
;	und zwar wird mit hilfe der Formel:
;
;		V = ADC + (a*V1)>>7 - V2
;		V2 = V1
;		V1 = V
;
;	Nach N_VAL = 205 Abtastungen werden dann aus den beiden werten von V1 und V2
;	unter Anwendungen der folgenden Formel die Leistung bei einer bestimmten Frequenz
;	berechnet.
;
;		V1 = V1 >> 8
;		V2 = V2 >> 8
;		Y = V1*V1 + V2*V2 - (V1*V2*a)>>10
;
;	Die Rechtsschiebe Operationen in beiden Formelen dienen zur Anpassung der Rechenergebnisse
;	auf das hier verwendete Interger Format von 16 Bit.
;
;	Es ergibt sich eine Rechnerische Bandbreite des Goertzel Filters
;	von FS(8Khz) / N_VAL(205) = 39 Hz was fuer die Detektion der DTMF Frequenzen
;	ausreichend ist.
;	Um nun festzustellen welche DTMF-Frequenz im Signalgemisch vorhanden ist
;	wird sich sowohl im Lowband wie auch im Highband die hoechste Amplitude gemerkt 
;	Liegt diese Amplitude im Lowband ueber den "AmpTrgLevelLow" so gilt diese Frequenz
;	als vorhanden im Eingangssignal das gleiche gilt fuer die Amplitude im Highband
;	die gegen den wert "AmpTrgLevelHigh" verglichen wird.
;	Wenn jetzt im Lowband und im Highband eine gueltige Frequenz detektiert wurde
;	wird je nachdem welche Ausgabe Option eingestellt das erkannte Zeichen ausgegeben.
;
;	Berechnung der Faktoren N,a
;	Den Faktor N der die Bandbreite des Goertzel Filter bestimmt wird folgendermassen
;	berechnet
;		N = FS/B
;		FS = Abtastfrequenz
;		B  = Bandbreite
;		N = 8000 / 39Hz = 205
;
;	Der Faktor a berechnet 	sich folgender massen
;		a = 2 * cos ((2*Pi*fk)/FS)
;	
;	Dabei ist fk die zu detektierende Frequenz (697,770,852,941) Lowband und
;	(1209,1336,1477,1633) Highband
;
;	da hier aber mit ganzzahl arithmetik gerechnet wird mus dieser
;	faktor a noch mit 128 skaliert werden
;	a = a * 128
;
;	Der Faktor sorgt dafuer das der Wertebereich von 16 moeglichst ausgenutz
;	wird und es keine ueberlauefe mit Vorzeichenwechsel gibt
;
;	Bei der Programmierung des Mikrokontrollers muss auf das richtige
;	Setzen der FUSE bits geachtet werden. Es muessen die FUSE fuer 
;	externen Reset und externer Oszillator > 10Mhz gesetzt sein.
;
;	Version 1.0 H.Lenzen  Oktober 2007
;
;****************************************************************************


		.include "tn2313def.inc"

		.equ	DISP = 0				;7 Segment ausgabe
		.equ	PARA = 1				;4 Bit Parallel ausgabe mit Strobe
		.equ	SERI = 0				;Serielle ausgabe mit 9600 Bd

		.equ	AmpTrgLevelLow = 0x40	;amplituden trigger schwelle Low Band
		.equ	AmpTrgLevelHigh = 0x40  ;amplituden trigger schwelle High Band
		.equ	MAXFREQ		= 8			;anzahl der frequenzen 
		.equ	T0_RELOAD   = 217		; 217 = 8012.8 Hz Sampling Frequenz

		.equ    RAMSTART	= $60		; Start On-Chip SRAM Location
		.equ	LASTOUT		= RAMSTART
		.equ	NVAL		= LASTOUT+1
		.equ	VARSTART	= NVAL+2
		.equ	Y_START		= VARSTART + MAXFREQ*4
		.equ	N_VAL		= 205

		.equ	BAUDRATE = 129			;9600 Baud bei 20 Mhz
		.equ	UCSRA_VAL = 0x00
		.equ	UCSRC_VAL = 0x06
		.equ	UCSRB_VAL = 0x08

		.equ	SEG_A = 0x01
		.equ	SEG_B = 0x02
		.equ	SEG_C = 0x04
		.equ	SEG_D = 0x08
		.equ	SEG_E = 0x10
		.equ	SEG_F = 0x20
		.equ	SEG_G = 0x40

		.equ	PAT_0 = SEG_A+SEG_B+SEG_C+SEG_D+SEG_E+SEG_F
		.equ	PAT_1 = SEG_B+SEG_C
		.equ	PAT_2 = SEG_A+SEG_B+SEG_G+SEG_E+SEG_D
		.equ	PAT_3 = SEG_A+SEG_B+SEG_C+SEG_D+SEG_G
		.equ	PAT_4 = SEG_B+SEG_C+SEG_G+SEG_F
		.equ	PAT_5 = SEG_A+SEG_F+SEG_G+SEG_C+SEG_D
		.equ	PAT_6 = SEG_A+SEG_F+SEG_G+SEG_C+SEG_D+SEG_E
		.equ	PAT_7 = SEG_A+SEG_B+SEG_C
		.equ	PAT_8 = SEG_A+SEG_B+SEG_C+SEG_D+SEG_E+SEG_F+SEG_G
		.equ	PAT_9 = SEG_A+SEG_B+SEG_C+SEG_D+SEG_F+SEG_G
		.equ	PAT_A = SEG_A+SEG_B+SEG_C+SEG_F+SEG_E+SEG_G
		.equ	PAT_B = SEG_C+SEG_D+SEG_E+SEG_F+SEG_G
		.equ	PAT_C = SEG_A+SEG_D+SEG_E+SEG_F
		.equ	PAT_D = SEG_B+SEG_C+SEG_D+SEG_E+SEG_G
		.equ	PAT_R = SEG_G+SEG_E
		.equ	PAT_S = SEG_A+SEG_B+SEG_G+SEG_F+SEG_E
			

		.def	mc16sL	=r16		;multiplicand low byte
		.def	mc16sH	=r17		;multiplicand high byte
		.def	mp16sL	=r18		;multiplier low byte
		.def	mp16sH	=r19		;multiplier high byte
		.def	m16s0	=r18		;result byte 0 (LSB)
		.def	m16s1	=r19		;result byte 1
		.def	m16s2	=r20		;result byte 2
		.def	m16s3	=r21		;result byte 3 (MSB)
		.def	mcnt16s	=r22		;loop counter


		rjmp	MAIN		; Reset Handler
		rjmp	EXT_INT0	; IRQ0 Handler
		rjmp	EXT_INT1	; IRQ0 Handler
		rjmp	TIM1_CAPT	; Timer1 Capture Handler
		rjmp	TIM1_COMP	; Timer1 Compare Handler
		rjmp	TIM1_OVF	; Timer1 Overflow Handler
		rjmp	TIM0_OVF	; Timer0 Overflow Handler
		rjmp	UART_RXC	; UART RX Complete Handler
		rjmp	UART_DRE	; UDR Empty Handler
		rjmp	UART_TXC	; UART TX Complete Handler
		rjmp	ANA_COMP	; Analog Comparator Handler


EXT_INT0:
EXT_INT1:
TIM1_CAPT:
TIM1_COMP:
TIM1_OVF:
UART_RXC:
UART_DRE:
UART_TXC:
ANA_COMP:
		reti


MAIN:
	; **** Initialize Stack Pointer ****
		ldi		R24, LOW(RAMEND)	; Stack Pointer Setup
		out		SPL, R24			; Stack Pointer Low Byte	
	; **** Turn off Watchdog timer ****
		ldi		R24, 0x18			; 
		out		WDTCR, R24			; 
		ldi		R24, 0x10			; 
		out		WDTCR, R24			; 

		; **** Initialize I/O Port ****
		ldi		R24, 0xfc			; Set PB2-7 as Output Mode
		out		DDRB, R24
		ldi		R24, 0x00			; DAC = 0 V out
		out		PORTB, R24

		.if		DISP
		ldi		R24,0x7f			;PD 0.. 6 -> output
		out		DDRD,R24
		.endif

		.if		PARA
		ldi		R24,0x7f			;PD 0.. 6 -> output
		out		DDRD,R24
		.endif

		.if		SERI
		ldi		R24,0x7c			;PD 2.. 6 -> output PD 0..1 Input
		out		DDRD,R24
		.endif

		ldi		R24, 0x02			; Enable Timer0 Overflow Interrupt
		out		TIMSK, R24
		ldi		R24, T0_RELOAD		; Timer 0 Reload
		out		TCNT0, R24
		ldi		R24, 0x03			; CK/64, Timer Start
		out		TCCR0, R24

		ldi		xl,low(VARSTART)
		ldi		xh,high(VARSTART)
		ldi		R16,MAXFREQ
		clr		r17
ClearLoop:
		st		X+,R17				;low byte loeschen  (V1)
		st		X+,R17				;high byte loeschen (V1)
		st		X+,R17				;V2
		st		X+,R17				;V2
		dec		r16
		brne	ClearLoop
		
		ldi		R16,low(N_VAL)		;n Counter setzen
		sts		NVAL,R16
		ldi		R16,high(N_VAL)
		sts		NVAL+1,R16

		ldi		r16,0xff
		sts		LASTOUT,r16			;clear Lastout

;******* Init Uart  (nur gueltig wenn Serielle Ausgabe ************
		.if		SERI
		ldi		R16,UCSRA_VAL		;9600 Bd,8 Databits,1 Stopbit,TX Enable einstellen
		out		UCSRA,R16
		ldi		R16,UCSRB_VAL
		out		UCSRB,R16
		ldi		R16,UCSRC_VAL
		out		UCSRC,R16
		ldi		R16,low(BAUDRATE)
		out		UBRRL,R16
		ldi		R16,high(BAUDRATE)
		out		UBRRH,R16
		.endif

		sei							; Enable Global Interrupt
forever:	
		rjmp	forever

;-----------------------------------------------------------------------------------------
;
;	SARADC
;	ADC Converter routine
;	ergebnis in R16(low),R17(high)
;	register R16,R17,R18
;	ca. 98 Clocks
;-----------------------------------------------------------------------------------------

SARADC:	ldi		R18,0x80			;probe bit maske
		clr		R16					;result register
		ldi		R17,6				;6 bit to convert
SARLOOP:
		or		R16,R18				;set probe bit
		out		PORTB,R16			;out to DAC
		nop
		nop
		nop
		nop
		nop
		nop
		nop
		nop							;500 nS warten fuer Comparator bei 16Mhz Takt
		sbic	ACSR,5				;DAC wert > analog wert ?
		eor		R16,R18				;wenn ja probe bit zurueck nehmen
		lsr		R18					;naechstes probe bit schieben
		subi	R17,1				;decrement Loop counter
		brne	SARLOOP				;alle durch ?
		lsr		R16
		lsr		R16					;ergebnis schieben
		subi	r16,32				;bereich 0..63 -> -32 .. +31
		clr		r17
		lsl		r16					;*2 = -64..+62
		brpl	SAR1
		com		r17					;vorzeichen erweiterung in r17
SAR1:	ret							;fertig wert in R16


;***************************************************************************
;*
;* "mpy16s" - 16x16 Bit Signed Multiplication
;*
;* This subroutine multiplies signed the two 16-bit register variables 
;* mp16sH:mp16sL and mc16sH:mc16sL.
;* The result is placed in m16s3:m16s2:m16s1:m16s0.
;* The routine is an implementation of Booth's algorithm. If all 32 bits
;* in the result are needed, avoid calling the routine with
;* -32768 ($8000) as multiplicand
;*  
;* Number of words	:16 + return
;* Number of cycles	:210/226 (Min/Max) + return
;* Low registers used	:None
;* High registers used  :7 (mp16sL,mp16sH,mc16sL/m16s0,mc16sH/m16s1,
;*			    m16s2,m16s3,mcnt16s)	
;*
;***************************************************************************

mpy16s:	clr		m16s3			;clear result byte 3
		sub		m16s2,m16s2		;clear result byte 2 and carry
		ldi		mcnt16s,16		;init loop counter
m16s_1:	brcc	m16s_2			;if carry (previous bit) set
		add		m16s2,mc16sL	;add multiplicand Low to result byte 2
		adc		m16s3,mc16sH	;add multiplicand High to result byte 3
m16s_2:	sbrc	mp16sL,0		;if current bit set
		sub		m16s2,mc16sL	;sub multiplicand Low from result byte 2
		sbrc	mp16sL,0		;if current bit set
		sbc		m16s3,mc16sH	;sub multiplicand High from result byte 3
		asr		m16s3			;shift right result and multiplier
		ror		m16s2
		ror		m16s1
		ror		m16s0
		dec		mcnt16s			;decrement counter
		brne	m16s_1			;if not done, loop more	
		ret

;-----------------------------------------------------------------------------------------
;
;	Timer 0 Overflow interrupt
;
;-----------------------------------------------------------------------------------------

TIM0_OVF:
		ldi		R24, T0_RELOAD		; Timer Count
		out		TCNT0, R24			; reload timer

;********** ADC wert holen ***********

		rcall	SARADC				;ADC in R16
		mov		r10,r16				;in R10,R11 sichern
		mov		r11,r17

;********** V = ADC + V1 * a - V2 berechnen ********

		ldi		yl,low(VARSTART)	;ptr auf variabeln bereich
		ldi		yh,high(VARSTART)
		ldi		zl,low(FAC_A*2)		;ptr auf factor A Frequenz
		ldi		zh,high(FAC_A*2)
		ldi		R24,MAXFREQ			;counter ueber alle frequenzen
FreqLoop1:
		mov		r2,r10				;adc wert laden
		mov		r3,r11
		lpm		r8,z+				;a laden
		lpm		r9,z+
		rcall	CalcInnerGoertzel
		adiw	Y,4					;ptr to next V1,V2
		dec		R24
		brne	FreqLoop1			;alle frequenzen durchgerechnet ?

;************* Test n >= N_VAL ******************

		lds		xl,NVAL				;n Laden
		lds		xh,NVAL+1
		sbiw	x,1					;n-1
		sts		NVAL,xl
		sts		NVAL+1,xh
		or		xl,xh				;n = 0?
		cpi		xl,0
		breq	L2
		rjmp	TIM0_END			;nein dann fertig
L2:		ldi		xl,low(N_VAL)
		ldi		xh,high(N_VAL)
		sts		NVAL,xl
		sts		NVAL+1,xh			;n neu laden


;************** Y = V1*V1 + V2*V2 - V1*V2*a ****************

		ldi		yl,low(VARSTART)	;ptr auf variabeln bereich
		ldi		yh,high(VARSTART)
		ldi		zl,low(FAC_A*2)		;ptr auf factor A
		ldi		zh,high(FAC_A*2)
		ldi		xl,low(Y_START)		;ptr auf die Y werte
		ldi		xh,high(Y_START)
		clr		r23					;loop counter
FreqLoop2:
		lpm		r8,z+				;a  laden
		lpm		r9,z+				;ptr to next a
		rcall	CalcOuterGoertzel	;Y berechnen
		adiw	y,4					;ptr to next V1,V2
		st		X+,r2				;Y-Wert speichern (low byte)
		st		X+,r3				;				  (high byte)

		inc		r23					;inc Loop counter
		cpi		r23,MAXFREQ			;alle frequenzen durch ?
		brlo	FreqLoop2			;nein dann naechste frequenz berechnen

;*************** suche den maximalen Y wert im Low Band ******************

		ldi		xl,low(Y_START)		;ptr auf die Y werte
		ldi		xh,high(Y_START)
		clr		r2					;clear maximum Low Band
		clr		r3
		clr		r4					;clear maximum high band
		clr		r5					
		clr		r16					;clear loop counter
		clr		r17					;Clear Freq nummer low
		clr		r18					;clear Freq nummer high

LowLoop:
		ld		r6,X+				;load Y wert
		ld		r7,X+
		cp		r6,r2
		cpc		r7,r3
		brlt	NextLow				;test  Y > OldY ?
		mov		r2,r6				;wenn ja dann neues Maximum setzen
		mov		r3,r7
		mov		r17,r16				;Frequenz nummer merken 1..4
		inc		r17
NextLow:
		inc		r16					;loop counter
		cpi		r16,MAXFREQ/2
		brlo	LowLoop
			
;************ suche den Maximalen Y Wert im High Band ********************

		clr		r16					;clear loop counter
HighLoop:
		ld		r6,X+
		ld		r7,X+
		cp		r6,r4
		cpc		r7,r5
		brlt	NextHigh			;test Y > Old Y?
		mov		r4,r6				;wenn ja dann neues Maximum merken
		mov		r5,r7
		mov		r18,r16
		inc		r18					;Frequenz nummer merken 1..4
NextHigh:
		inc		r16					;loop counter
		cpi		r16,MAXFREQ/2
		brlo	HighLoop

;************ Maximum High Low gefunden ****************************

FoundHighLow:
		ldi		r19,low(AmpTrgLevelLow)	;test ob die Trigger schwelle erreicht wird
		ldi		r20,high(AmpTrgLevelLow)
		cp		r2,r19
		cpc		r3,r20
		brge	LowOk
		clr		r17						;Low nicht Ok
LowOk:	
		ldi		r19,low(AmpTrgLevelHigh);test ob die Trigger schwelle erreicht wird
		ldi		r20,high(AmpTrgLevelHigh)
		cp		r4,r19
		cpc		r5,r20
		brge	HighOk
		clr		r18						;High nicht Ok
HighOk:

;************ Serielle ASCII ausgabe ***********************************

		.if		SERI
		cpi		r17,0					;lowband ok ?
		breq	NoTone
		cpi		r18,0					;highband ok ?
		breq	NoTone
		dec		r17
		dec		r18
		lsl		R18						;Highnummer * 4
		lsl		R18
		add		r17,r18
		ldi		zl,low(TAB_ASCII*2)		;hole ascii code aus tabelle
		ldi		zh,high(TAB_ASCII*2)
		clr		r18
		add		zl,r17
		adc		zh,r18
		lpm		
		lds		r1,LASTOUT
		cp		r0,r1					;neuer Tone erkannt
		brne	Serout1
		rjmp	TIM0_END				;nein dann fertig

Serout1:
		sbis	UCSRA,UDRE				;dann gib das zeichen aus
		rjmp	Serout1
		out		UDR,r0
		sts		LASTOUT,r0				;zeichen merken
		rjmp	TIM0_END

NoTone:	ldi		r16,0xff				;clear Lastout
		sts		LASTOUT,r16
		rjmp	TIM0_END
		.endif

;************ 4-Bit Parallel ausgabe ***********************************
		.if		PARA
		cpi		r17,0					;lowband ok ?
		breq	NoTone
		cpi		r18,0					;highband ok ?
		breq	NoTone
		dec		r17
		dec		r18
		lsl		R18						;Highnummer * 4
		lsl		R18
		add		r17,r18
		ldi		zl,low(TAB_PARA*2)		;hole code aus tabelle
		ldi		zh,high(TAB_PARA*2)
		clr		r18
		add		zl,r17
		adc		zh,r18
		lpm		
		lds		r1,LASTOUT
		cp		r0,r1					;neuer Tone erkannt ?
		brne	Para1					;ja
		rjmp	TIM0_END				;nein dann fertig

Para1:	sts		LASTOUT,r0				;zeichen als LASTOUT merken
		lsl		r0
		lsl		r0
		lsl		r0						;bits auf position schieben
		out		PORTD,r0				;daten ausgeben
		nop								;Delay
		nop
		nop
		nop	
		ldi		r16,0x04
		or		r0,r16					;Strobe ausgeben
		out		PORTD,r0
		com		r16
		and		r0,r16
		nop
		nop
		nop
		nop								;Delay
		out		PORTD,r0				;Strobe wieder wegnehmen
		rjmp	TIM0_END

NoTone:	ldi		r16,0xff
		sts		LASTOUT,R16				;clear Lastout
		rjmp	TIM0_END
		.endif

;************ 7 Segment ausgabe ****************************************

		.if		DISP
		cpi		r17,0					;Lowband Ok ?
		breq	NoTone
		cpi		r18,0					;Highband ok ?
		breq	NoTone
		dec		r17
		dec		r18
		lsl		R18						;Highnummer * 4
		lsl		R18
		add		r17,r18

		ldi		zl,low(TAB_7*2)
		ldi		zh,high(TAB_7*2)
		clr		r18
		add		zl,r17
		adc		zh,r18
		lpm		
		com		r0
		out		PORTD,r0
		rjmp	TIM0_END
NoTone:
		ldi		r17,0xff				;Display aus
		out		PORTD,r17
		rjmp	TIM0_END
		.endif


TIM0_END:
		reti

;-------------------------------------------------------------------------------------
;
;	CalcInnerGoertzel
;	in Y ptr auf V1+V2
;	in R2,R3 den adc wert
;	in R8,R9 den faktor a
;
;	berechnet die innere goertzel schleife
;
;	v(r2,r3) = adc(r2,r3) - V2(r6,r7) + a(r8,r9)*v1(r4,r5) >> 8
;	V2 = V1
;	V1 = v;
;	gebrauchte register (R2,R3,R6,R7,R8,R9)
;
;-------------------------------------------------------------------------------------

CalcInnerGoertzel:

		ldd		mp16sl,Y+0			;load v1
		ldd		mp16sh,Y+1
		ldd		r6,Y+2
		ldd		r7,Y+3				;load v2
		std		Y+2,mp16sl
		std		Y+3,mp16sh			;V2 = V1
		sub		r2,r6
		sbc		r3,r7				;v(r2,r3) = adc(r2,r3) - V2(r6,r7)
		mov		mc16sl,r8
		mov		mc16sh,r9
		rcall	mpy16s				;a(r8,r9) * V1(r4,r5)
		lsl		m16s0
		rol		m16s1
		rol		m16s2				; >> 7
		add		r2,m16s1
		adc		r3,m16s2			;v(r2,r3) = adc(r2,r3) - V2(r6,r7) + a(r8,r9)*v1(r4,r5) >> 7
		std		Y+0,r2
		std		Y+1,r3				;V1 = v
		ret

;-------------------------------------------------------------------------------------
;
;	CalcOuterGoertzel
;	berechnet die aussere Goertzel schleife
;
;	Y = (v1>>8)*(v1>>8)+(v2>>8)*(v2>8)-(v1*v2*a) >> 11
;	V1 = 0
;	V2 = 0
;	in Y ptr auf V1+V2
;	in R8,R9 den faktor a
;	Y = R2,R3,R4,R5
;
;-------------------------------------------------------------------------------------
CalcOuterGoertzel:
		ldd		mp16sl,Y+1		;v1 >> 8
		mov		mc16sl,mp16sl
		clr		mp16sh
		clr		mc16sh
		bst		mp16sl,7
		brtc	CalcOuter1
		com		mp16sh
		com		mc16sh
CalcOuter1:
		mov		r6,mp16sl
		mov		r7,mp16sh		;save v1 >> 8
		rcall	mpy16s			;Y = (v1>>8)*(v1>>8)
		mov		r2,m16s0
		mov		r3,m16s1
		mov		r4,m16s2
		mov		r5,m16s3		;store in R2,R3,R4,R5

		ldd		mp16sl,Y+3		;v2 >> 8
		mov		mc16sl,mp16sl
		clr		mp16sh
		clr		mc16sh
		bst		mp16sl,7
		brtc	CalcOuter2
		com		mp16sh
		com		mc16sh
CalcOuter2:
		mov		r12,mp16sl
		mov		r13,mp16sh		;save v2 >> 8
		rcall	mpy16s			;Y = (v2>>8)*(v2>>8)
			
		add		r2,m16s0
		adc		r3,m16s1
		adc		r4,m16s2
		adc		r5,m16s3		;Y = v1*v1+v2*v2

		mov		mp16sl,r6		;load v1 >> 8
		mov		mp16sh,r7
		mov		mc16sl,r12
		mov		mc16sh,r13		;load v2 >> 8
		rcall	mpy16s			;v1 * v2
		mov		mc16sl,r8
		mov		mc16sh,r9		;mp16sl,mp16sh = m16s0,m16s1
		rcall	mpy16s			;v1*v2*a
		ror		m16s3
		ror		m16s2
		ror		m16s1
		ror		m16s3
		ror		m16s2
		ror		m16s1			;v1*v2*a >>10
		sub		r2,m16s1
		sbc		r3,m16s2
		sbc		r4,m16s3
		clr		m16s0
		sbc		r5,m16s0		;Y = v1*v1+v2*v2-(v1*v2*a)>>10
		std		Y+0,m16s0
		std		Y+1,m16s0
		std		Y+2,m16s0
		std		Y+3,m16s0		;V1 = V2 = 0
		ret



;--------------------------------------------------------------------------
;	
;	Die ASCII Tabelle fuer die Serielle Ausgabe
;
;--------------------------------------------------------------------------
		.if		SERI
TAB_ASCII:
		.db		'1','4','7','*','2','5','8','0'
		.db		'3','6','9','#','A','B','C','D'
		.endif

;--------------------------------------------------------------------------
;
;	Die 7-Segment Tabelle
;
;--------------------------------------------------------------------------
		.if	DISP
TAB_7:	.db		PAT_1,PAT_4,PAT_7,PAT_S,PAT_2,PAT_5,PAT_8,PAT_0
		.db		PAT_3,PAT_6,PAT_9,PAT_R,PAT_A,PAT_B,PAT_C,PAT_D
		.endif
;--------------------------------------------------------------------------
;
;	Die Tabelle fuer die parallele ausgabe
;
;--------------------------------------------------------------------------
		.if	PARA
TAB_PARA:
		.db		0x01,0x04,0x07,0x0e,0x02,0x05,0x08,0x00
		.db		0x03,0x06,0x09,0x0f,0x0a,0x0b,0x0c,0x0d
		.endif
;--------------------------------------------------------------------------
;
;	Die faktoren 'a'
;
;--------------------------------------------------------------------------

FAC_A:		.dw		219,211,201,189,149,128,103,73
		


		.org		0x400
		.dw			0xefff				;Fuse wert fuer HED Chip Programmer
